# -*- coding: utf-8 -*-
""" Tests for the plugin.

Includes both tests written in unittest style (test_cli.py) and tests written
in pytest style (test_calculations.py).
"""
import os

TEST_DIR = os.path.dirname(os.path.realpath(__file__))
